# -*- coding: utf8 -*-
"""Manager for entry point based plugins. Main client facing API"""
from __future__ import print_function
import re

import six

from reentry.jsonbackend import JsonBackend

__all__ = ['PluginManager', 'get_default_backend', 'DEFAULT_MANAGER']

IGNORE_BY_DEFAULT = {
    'console_scripts', 'gui_scripts', 'distutils.commands', 'distutils.setup_keywords', 'setuptools.installation',
    'setuptools.file_finders', 'egg_info.writers'
}


def clean_map(entry_point_map, exceptions=None):
    """Extract entry points that are clearly not for plugins unless excepted."""
    ep_map = entry_point_map.copy()
    ignore_set = IGNORE_BY_DEFAULT.difference(set(exceptions or []))
    for ignore_group in ignore_set:
        ep_map.pop(ignore_group, None)
    return ep_map


def get_default_backend():
    from reentry.config import get_datafile
    return JsonBackend(datafile=get_datafile())


class PluginManager(object):
    """Manage a set of cached entry points"""

    def __init__(self, backend=get_default_backend(), scan_for_not_found=True):
        self._backend = backend
        self._scan_for_not_found = scan_for_not_found

    def iter_entry_points(self, group, name=None):
        """
        Iterate over all registered entry points in `group`, or if `name` is
        given, only the ones matching `name`.

        If no entry point is found in the group (or none called `name`), a scan
        is triggered, which may take a while. This behaviour can be configured
        when creating the plugin manager.

        The backend may only import pkg_resources if any of the entry points
        contain extras requirements.
        """
        if self._scan_for_not_found:
            if not self.get_entry_map(groups=[group]):
                self.scan(groups=[group])
            elif name and name not in self.get_entry_map(groups=[group], ep_names=[name]).get(group, {}).keys():
                self.scan(groups=[group])
        for entry_point in self._backend.iter_group(group):
            if name:
                if name in entry_point.name:
                    yield entry_point
            else:
                yield entry_point

    def get_entry_map(self, dist_names=None, groups=None, ep_names=None):
        """
        return the entry point map for `group` or the whole map for `dist`

        The backend may only load pkg_resources if any of the entry points contain extras requirements
        """
        return self._backend.get_map(dist=dist_names, group=groups, name=ep_names)

    def get_dist_map(self, dist=None):
        """Get a map of entry points sorted by distribution."""
        return self._backend.get_dist_map(dist=dist)

    @staticmethod
    def format_map(entry_point_map, indent=1):
        tabs = '\t' * indent
        newl = '\n' + tabs
        return tabs + newl.join(['{} -> {}'.format(dname, dmap) for dname, dmap in entry_point_map.items()])

    def register(self, distribution):
        """
        Registers the distribution's entry points with the backend.

        The backend may load pkg_resources to resolve the distribution name

        Takes either a string or a Distribution object as passed by setuptools to hooks during install.
        """
        dist_name, entry_point_map = self._backend.scan_dist(distribution)
        entry_point_map = clean_map(entry_point_map)
        self._backend.write_dist_map(dist_name, entry_point_map)
        return dist_name, entry_point_map

    def scan(self, groups=None, group_re=None, commit=True, delete=True):  # pylint: disable=too-many-branches
        """Walk through all distributions available and register entry points.

        Note: This imports pkg_resources.

        :param groups: a list of group names to register entry points for.
            If None, registers all entry points found.
        :param group_re: a regular expression for group names.
            Groups matched by the regular expression are appended to `groups`
        :type group_re: str or a compiled expression from re.compile
        :param commit: If False, performs just a dry run
        :param delete: If False, append to existing entry point map
        """
        import pkg_resources as pr
        pr_env = pr.AvailableDistributions()
        pr_env.scan()

        groups = groups or []

        if group_re:
            all_groups = self.scan_all_group_names()
            if isinstance(group_re, six.string_types):
                group_re = re.compile(group_re)
            groups.extend({group for group in all_groups if group_re.match(group)})

        if delete:
            full_map = {}

            if commit:
                if groups:
                    for group in groups:
                        self._backend.rm_group(group)
                else:
                    self.clear()

        else:
            full_map = self._backend.epmap.copy()

        # ~ for dists in pr_env._distmap.values():  # pylint: disable=protected-access
        for dist in pr_env:
            # note: in pip 19, the *installing* distribution is part of pr_env but
            # pkg_resources.get_distribution() fails on it
            try:
                dname, emap = self._backend.scan_dist(dist)
            except pr.DistributionNotFound:
                continue

            dmap = full_map.get(dname, {})
            if groups:
                new_dmap = {k: v for k, v in six.iteritems(emap) if k in groups}
                dmap.update(new_dmap)
            else:
                dmap.update(emap)

            # extract entry points that are reserved for other purposes unless excepted
            dmap = clean_map(dmap, exceptions=groups)

            if commit:
                self._backend.write_dist_map(dname, entry_point_map=dmap)
            full_map[dname] = [dmap]

        return full_map

    def scan_all_group_names(self):
        """Use `pkg_resources` to get a set of all available (not only cached) groups."""
        import pkg_resources as pr
        pr_env = pr.AvailableDistributions()
        pr_env.scan()
        all_groups = set()
        for dist_name in pr_env:
            _, dmap = self._backend.scan_dist(dist_name)
            all_groups.update(dmap.keys())
        return all_groups

    def unregister(self, distname):
        """
        unregisters the distribution's entry points with the backend

        The backend may load pkg_resources to resolve the distribution name
        """
        self._backend.rm_dist(distname)

    def clear(self):
        """Clear entry point map."""
        self._backend.clear()

    @property
    def distribution_names(self):
        return self._backend.get_dist_names()


DEFAULT_MANAGER = PluginManager()



# pylint: disable=unused-import,redefined-outer-name
"""Unit tests for config tools."""
try:
    # prefer the backport for Python <3.5
    from pathlib2 import Path
except ImportError:
    from pathlib import Path

import os
import pytest  # pylint: disable=unused-import
import six
from six.moves import configparser

from reentry import config


def test_find_config():
    config_file = config.find_config()
    assert isinstance(config_file, Path)


def test_make_config_parser():
    """Make sure we get the right configparser type."""
    parser = config.make_config_parser()
    if six.PY2:
        assert isinstance(parser, configparser.SafeConfigParser)
    else:
        assert isinstance(parser, configparser.ConfigParser)


def _check_config_valid(parser, expected_filename=None):
    """
    Perform validations for a given config.
    If expected_filename is given, check its value too.
    """
    if six.PY2:
        assert isinstance(parser, configparser.SafeConfigParser)
    else:
        assert isinstance(parser, configparser.ConfigParser)

    assert parser.get('general', 'datadir')
    assert parser.get('general', 'data_filename')
    if expected_filename:
        assert parser.get('general', 'data_filename') == expected_filename


def test_get_config():
    """Make sure the configparser gets created correctly."""
    parser = config.get_config()
    _check_config_valid(parser)


def test_get_config_with_env_var():
    """Make sure the configparser gets created correctly when REENTRY_DATA_FILENAME is set."""
    data_filename = 'entrypoints'
    os.environ['REENTRY_DATA_FILENAME'] = data_filename
    parser = config.get_config()
    os.environ.pop('REENTRY_DATA_FILENAME')

    _check_config_valid(parser, data_filename)


def test_get_datafile():
    datafile = Path(config.get_datafile())

    assert datafile.is_file()
    assert datafile.exists()



# -*- coding: utf8 -*-
"""Abstract base class for backends"""
import abc
import six


class BackendInterface(object):
    """
    Backend interface, subclass to implement a concrete backend.


    All methods without a method body need to be implemented in a backend.
    """
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def get_map(self, dist=None, group=None, name=None):
        """
        get a map of entry points, filtered by

        :param dist: distribution name or sequence of distribution names
        :param groups: single group name or sequence of group names
        :param name: entry point name pattern or sequence of name patterns

        The map is structured as follows::

            map = {
                    '<group>': {
                        '<entrypoint name>: <EntryPoint instance>
                        ...
                    },
                    ...
                },
        """

    @abc.abstractmethod
    def iter_group(self, group):
        """
        returns a list of entry points for the given group name
        """

    @abc.abstractmethod
    def get_group_names(self):
        """
        returns a list of group names
        """

    @abc.abstractmethod
    def get_dist_names(self):
        """
        returns a list of distribution names
        """

    @abc.abstractmethod
    def get_dist_map(self, dist):
        """
        returns a map {group:[entry_points, ...], ...} for the given dist name
        """

    @abc.abstractmethod
    def scan_st_dist(self, dist):
        """Scan a distribution given by a name, empty by default."""

    @abc.abstractmethod
    def scan_install_dist(self, dist):
        """Add an incomplete distribution as passed by setuptools during it's installation."""

    def scan_dist(self, distribution):
        """Take a distribution's project name, add the distribution."""
        if isinstance(distribution, six.string_types):
            dist_name, entry_point_map = self.scan_st_dist(distribution)
        else:
            dist_name, entry_point_map = self.scan_install_dist(distribution)
        return dist_name, entry_point_map

    @abc.abstractmethod
    def write_dist_map(self, distname, entry_point_map=None):
        """Write a distribution given the name and entry point map"""

    @abc.abstractmethod
    def rm_dist(self, distname):
        """
        removes a distribution completely
        """

    @abc.abstractmethod
    def clear(self):
        """Clears all stored entry points"""

    @staticmethod
    def pr_dist_map(dist):
        dname = dist.project_name
        epmap = dist.get_entry_map()
        return dname, epmap

    @staticmethod
    def pr_dist_from_name(distname):
        from pkg_resources import get_distribution
        dist = get_distribution(distname)
        return dist

    @abc.abstractproperty
    def epmap(self):
        """Full map {distribution: {group: [{name: entry_point}]}}."""



# -*- coding: utf8 -*-
"""
Define a setuptools extension.

Usage::

    setup(
        ...,
        reentry_register = True,
        ...
    )
"""
from __future__ import print_function
import sys

from reentry.config import get_datafile


def is_bool(value):
    """Return True if `value` is a boolean."""
    return bool(value) == value


def register_dist(dist, attr, value):
    """If value is True, register the distribution's entry points in reentrys storage."""
    from distutils.errors import DistutilsSetupError  # pylint: disable=import-error,no-name-in-module
    # assert is boolean
    if not is_bool(value):
        raise DistutilsSetupError('{} must be a boolean, got {}'.format(attr, value))

    if value:
        print('[ REENTRY ] registering entry points with reentry...', file=sys.stderr)
        from reentry import manager
        dist_name, entry_point_map = manager.register(dist)
        print('[ REENTRY ] ... registered to {}'.format(get_datafile()), file=sys.stderr)
        print('[ REENTRY ] Following entrypoints were registered\n', file=sys.stderr)
        print(manager.format_map({dist_name: entry_point_map}), file=sys.stderr)
        print('[ REENTRY ] Current entry point map at {}:'.format(get_datafile()), file=sys.stderr)
        print(manager.format_map(manager.get_dist_map()), file=sys.stderr)


def ensure_list(value, attr):
    """raise an error if `value` is not a list"""
    from distutils.errors import DistutilsSetupError  # pylint: disable=import-error,no-name-in-module
    if not isinstance(value, list):
        raise DistutilsSetupError('{} must be a list, got {}'.format(attr, value.__class__))



"""Find and read user settings."""
import os
import sys
import hashlib
import platform
try:
    # prefer the backport for Python <3.5
    from pathlib2 import Path
except ImportError:
    from pathlib import Path

import six
from six.moves import configparser

__all__ = ['find_config', 'get_config', 'get_datafile']


def _get_default_config_dir():
    return Path(os.getenv('XDG_CONFIG_HOME', '~/.config')).expanduser().joinpath('reentry')


def find_config():
    """
    Search for a config file in the following places and order:

        * <HOME>/.reentryrc
        * <HOME>/.config/reentry/config
    """
    rc_file = Path.home().joinpath('.reentryrc')
    config_file = _get_default_config_dir().joinpath('config')
    # pylint: disable=no-else-return
    if rc_file.exists():  # pylint: disable=no-member
        return rc_file
    elif config_file.exists():  # pylint: disable=no-member
        return config_file
    return rc_file


def make_config_parser(*args, **kwargs):
    """Get the correct ConfigParser class depending on python version."""
    # pylint: disable=no-else-return
    if six.PY2:
        return configparser.SafeConfigParser(*args, **kwargs)
    elif six.PY3:
        return configparser.ConfigParser(*args, **kwargs)
    return None


def get_config(config_file_name=str(find_config())):
    """Create config parser with defaults and read in the config file."""
    default_config_dir = _get_default_config_dir()
    default_config_values = {'datadir': str(default_config_dir.joinpath('data')), 'data_filename': hashed_data_file_name()}
    parser = make_config_parser(default_config_values)
    parser.add_section('general')
    parser.read([config_file_name])

    env_datadir = os.getenv('REENTRY_DATADIR')
    if env_datadir:
        env_datadir_path = Path(env_datadir)
        if env_datadir_path.exists() and not env_datadir_path.is_dir():  # pylint: disable=no-member
            raise ValueError('environment variable $REENTRY_DATADIR={} exists, but is not a directory'.format(env_datadir))
        parser.set('general', 'datadir', str(env_datadir_path))

    env_data_filename = os.getenv('REENTRY_DATA_FILENAME')
    if env_data_filename:
        parser.set('general', 'data_filename', env_data_filename)

    return parser


def hashed_data_file_name():
    """Find the path to the reentry executable and mangle it into a file name."""

    fname = 'u{bin_dir}_{impl}-{ver}'.format(bin_dir=Path(sys.executable).resolve().parent,
                                             impl=platform.python_implementation(),
                                             ver=platform.python_version())

    path_hash = hashlib.sha256(fname.encode('utf-8'))
    return path_hash.hexdigest()


def get_datafile():
    """Create the path to the data file used to store entry points."""
    config = get_config()

    pkg_path_filename = config.get('general', 'data_filename')
    datafile = Path(config.get('general', 'datadir')).joinpath(pkg_path_filename)
    if not datafile.exists():  # pylint: disable=no-member
        datafile.parent.mkdir(parents=True, exist_ok=True)
        datafile.write_text(u'{}')

    return str(datafile)
